
import numpy as np
from rkstiff.solver import StiffSolverAS,StiffSolverCS

def phi1(z):
    """ Computes RKETD psi-function of the first order. 
    INPUTS
        z - real or complex-valued input array 
    OUTPUT
        return (exp(z)-1)/z  -> real or complex-valued RKETD function of first order
    """
    return (np.exp(z)-1)/z

def phi2(z):
    """ Computes RKETD psi-function of the second order. 
    INPUTS
        z - real or complex-valued input array 
    OUTPUT
        return 2!*(exp(z)-1-z/2)/z^2  -> real or complex-valued RKETD function of second order
    """
    return 2*(np.exp(z)-1-z)/z**2

def phi3(z):
    """ Computes RKETD psi-function of the third order. 
    INPUTS
        z - real or complex-valued input array 
    OUTPUT
        return 3!*(exp(z)-1-z/2-z^3/6)/z^3  -> real or complex-valued RKETD function of third order
    """
    return 6*(np.exp(z)-1-z-z**2/2)/z**3

class ETDAS(StiffSolverAS):
    """
    Class template for an RKETD adaptive-step solver. Adds several ETD specific parameters
    to the StiffSolverAS class.

    ATTRIBUTES
    ----------

    modecutoff : float
       psi RKETD functions are numerically inaccurate near zero when computed directly; when |z| < modecutoff use a
       contour integral (more computationally expensive method) to compute these values
    contour_points : int
        number of points to use in contour integral when evaluating small input values to the psi RKETD functions
    contour_radius : float
        radius size of a circular contour integral to be used to evaluate small input values to the psi RKETD functions

    """

    def __init__(self,linop,NLfunc,epsilon=1e-4,incrF = 1.25, decrF = 0.85, safetyF = 0.8,\
            adapt_cutoff = 0.01, minh = 1e-16, modecutoff = 0.01, contour_points = 32,\
            contour_radius = 1.0):
        super().__init__(linop,NLfunc,epsilon=epsilon,incrF=incrF,decrF=decrF,\
                safetyF=safetyF,adapt_cutoff=adapt_cutoff,minh=minh)
        self.modecutoff = modecutoff
        if (self.modecutoff > 1.0) or (self.modecutoff <= 0):
            raise ValueError('modecutoff must be between 0.0 and 1.0 but is {}'.format(self.modecutoff))
        self.contour_points = contour_points
        if not isinstance(self.contour_points,int):
            raise TypeError('contour_points must be an integer but is {}'.format(self.contour_points))
        if self.contour_points <= 1:
            raise ValueError('contour_points must be an integer greater than 1 but is {}'.format(self.contour_points))

        self.contour_radius = contour_radius
        if self.contour_radius <= 0:
            raise ValueError('contour_radius must greater than 0 but is {}'.format(self.contour_radius))
        self._h_coeff = None
        

class ETDCS(StiffSolverCS):
    """
    Class template for an RKETD constant-step solver. Adds several ETD specific parameters
    to the StiffSolverCS class.

    ATTRIBUTES
    ----------

    modecutoff : float
       psi RKETD functions are numerically inaccurate near zero when computed directly; when |z| < modecutoff use a
       contour integral (more computationally expensive method) to compute these values
    contour_points : int
        number of points to use in contour integral when evaluating small input values to the psi RKETD functions
    contour_radius : float
        radius size of a circular contour integral to be used to evaluate small input values to the psi RKETD functions

    """

    def __init__(self,linop,NLfunc,modecutoff = 0.01,contour_points = 32,contour_radius = 1.0):
        super().__init__(linop,NLfunc)
        self.modecutoff = modecutoff
        if (self.modecutoff > 1.0) or (self.modecutoff <= 0):
            raise ValueError('modecutoff must be between 0.0 and 1.0 but is {}'.format(self.modecutoff))
        self.contour_points = contour_points
        if not isinstance(self.contour_points,int):
            raise TypeError('contour_points must be an integer but is {}'.format(self.contour_points))
        if self.contour_points <= 1:
            raise ValueError('contour_points must be an integer greater than 1 but is {}'.format(self.contour_points))
        self.contour_radius = contour_radius
        if self.contour_radius <= 0:
            raise ValueError('contour_radius must greater than 0 but is {}'.format(self.contour_radius))
        self._h_coeff = None








from rkstiff.etd import ETDCS,phi1,phi2,phi3
import numpy as np
from typing import Callable
from scipy.linalg import expm

class _ETD4_Diagonal:
    """ ETD4 diagonal system strategy for ETD4 solver """
    def __init__(self,linop,NLfunc,contourM,contourR,modecutoff):
        self.linop = linop
        self.NLfunc = NLfunc
        self.M = contourM
        self.R = contourR
        self.modecutoff = modecutoff
        
        N = linop.shape[0]
        self._EL, self._EL2 = [np.zeros(N,dtype=np.complex128) for _ in range(2)]
        self._a21, self._a31, self._a32, self._a41, self._a43 = [np.zeros(N,dtype=np.complex128) for _ in range(5)]   
        self._b1, self._b2, self._b4 = [np.zeros(N,dtype=np.complex128) for _ in range(3)]   
        self._NL1, self._NL2, self._NL3,self._NL4 = [np.zeros(N,dtype=np.complex128) for _ in range(4)]
        self._k = np.zeros(N,dtype=np.complex128)
        
    def _updateCoeffs(self,h):
        z = h*self.linop
        self._EL = np.exp(z)
        self._EL2 = np.exp(z/2)
      
        smallmode_idx = np.abs(z) < self.modecutoff
        zb = z[~smallmode_idx] # z big
        # compute big mode coeffs
        phi1_12 = h*phi1(zb/2)
        phi2_12 = h*phi2(zb/2)
        phi1_1 = h*phi1(zb)
        phi2_1 = h*phi2(zb)
        phi3_1 = h*phi3(zb)
        
        self._a21[~smallmode_idx] = 0.5*phi1_12
        self._a31[~smallmode_idx] = 0.5*(phi1_12-phi2_12)
        self._a32[~smallmode_idx] = 0.5*phi2_12
        self._a41[~smallmode_idx] = phi1_1 - phi2_1
        self._a43[~smallmode_idx] = phi2_1
        self._b1[~smallmode_idx] = phi1_1 - (3.0/2)*phi2_1+(2.0/3)*phi3_1
        self._b2[~smallmode_idx] = phi2_1 - (2.0/3)*phi3_1
        self._b4[~smallmode_idx] = -(1.0/2)*phi2_1 + (2.0/3)*phi3_1
        
        # compute small mode coeffs
        zs = z[smallmode_idx] # z small
        r = self.R*np.exp(2j*np.pi*np.arange(0.5,self.M)/self.M)
        rr, zz = np.meshgrid(r,zs)
        Z = zz+rr
        
        phi1_12 = h*np.sum(phi1(Z/2),axis=1)/self.M
        phi2_12 = h*np.sum(phi2(Z/2),axis=1)/self.M
        phi1_1 = h*np.sum(phi1(Z),axis=1)/self.M
        phi2_1 = h*np.sum(phi2(Z),axis=1)/self.M
        phi3_1 = h*np.sum(phi3(Z),axis=1)/self.M
        
        self._a21[smallmode_idx] = 0.5*phi1_12
        self._a31[smallmode_idx] = 0.5*(phi1_12-phi2_12)
        self._a32[smallmode_idx] = 0.5*phi2_12
        self._a41[smallmode_idx] = phi1_1 - phi2_1
        self._a43[smallmode_idx] = phi2_1
        self._b1[smallmode_idx] = phi1_1 - (3.0/2)*phi2_1+(2.0/3)*phi3_1
        self._b2[smallmode_idx] = phi2_1 - (2.0/3)*phi3_1
        self._b4[smallmode_idx] = -(1.0/2)*phi2_1 + (2.0/3)*phi3_1
    
    def _N1_init(self,u):
        self._NL1 = self.NLfunc(u)
        
    def _updateStages(self,u):        
        # Use First is same as last principle (FSAL) 
        self._k = self._EL2*u + self._a21*self._NL1
        self._NL2 = self.NLfunc(self._k)
        self._k = self._EL2*u + self._a31*self._NL1 + self._a32*self._NL2
        self._NL3 = self.NLfunc(self._k)
        self._k = self._EL*u + self._a41*self._NL1 + self._a43*self._NL3
        self._NL4 = self.NLfunc(self._k)
        self._k = self._EL*u + self._b1*self._NL1 + self._b2*(self._NL2+self._NL3) \
                + self._b4*self._NL4
        self._NL1 = self.NLfunc(self._k)
        return self._k

class _ETD4_NonDiagonal:
    """ ETD4 non-diagonal system strategy for ETD4 solver """
    def __init__(self,linop,NLfunc,contourM,contourR):
        self.linop = linop
        self.NLfunc = NLfunc
        self.M = contourM
        self.R = contourR
        
        N = linop.shape[0]
        self._EL, self._EL2 = [np.zeros(shape=linop.shape,dtype=np.complex128) for _ in range(2)]
        self._a21, self._a31, self._a32, self._a41, self._a43 = [\
                np.zeros(shape=linop.shape,dtype=np.complex128) for _ in range(5)]   
        self._b1,self._b2, self._b4 = [np.zeros(shape=linop.shape,dtype=np.complex128) for _ in range(3)]   
        self._NL1, self._NL2, self._NL3,self._NL4 = [np.zeros(N,dtype=np.complex128) for _ in range(4)]
        self._k = np.zeros(N,dtype=np.complex128)
        
    def _updateCoeffs(self,h):   
        z = h*self.linop
        self._EL = expm(z)
        self._EL2 = expm(z/2)

        contour_points = self.R*np.exp(2j*np.pi*np.arange(0.5,self.M)/self.M)
        
        phi1_12, phi2_12,phi1_1,phi2_1,phi3_1 = [np.zeros(shape=self.linop.shape,dtype=np.complex128) for _ in range(5)]
        for point in contour_points:
            Q = np.linalg.inv(point*np.eye(*self.linop.shape)-z)
            Q2 = np.linalg.inv(point*np.eye(*self.linop.shape)-z/2)
            phi1_12 += point*phi1(point)*Q2/self.M
            phi2_12 += point*phi2(point)*Q2/self.M
            phi1_1 += point*phi1(point)*Q/self.M
            phi2_1 += point*phi2(point)*Q/self.M
            phi3_1 += point*phi3(point)*Q/self.M  
        
        self._a21 = 0.5*h*phi1_12
        self._a31 = 0.5*h*(phi1_12-phi2_12)
        self._a32 = 0.5*h*phi2_12
        self._a41 = h*(phi1_1 - phi2_1)
        self._a43 = h*phi2_1
        self._b1 = h*(phi1_1 - (3.0/2)*phi2_1+(2.0/3)*phi3_1)
        self._b2 = h*(phi2_1 - (2.0/3)*phi3_1)
        self._b4 = h*(-(1.0/2)*phi2_1 + (2.0/3)*phi3_1)
        
    def _N1_init(self,u):
        self._NL1 = self.NLfunc(u)
        
    def _updateStages(self,u):        
        self._k = self._EL2.dot(u) + self._a21.dot(self._NL1)
        self._NL2 = self.NLfunc(self._k)
        self._k = self._EL2.dot(u) + self._a31.dot(self._NL1) + self._a32.dot(self._NL2)
        self._NL3 = self.NLfunc(self._k)
        self._k = self._EL.dot(u) + self._a41.dot(self._NL1) + self._a43.dot(self._NL3)
        self._NL4 = self.NLfunc(self._k)
        self._k = self._EL.dot(u) + self._b1.dot(self._NL1) + self._b2.dot(self._NL2+self._NL3) \
                + self._b4.dot(self._NL4)
        self._NL1 = self.NLfunc(self._k) # Use First is same as last principle (FSAL) 
        return self._k

class ETD4(ETDCS):
    """
    Exponential time-differencing constant step solver of 4th order (Krogstad)

    ATTRIBUTES
    __________

    linop : np.array
    NLfunc : function
    t : time-array stored with evolve function call 
    u : output-array stored with evolve function call 
    logs : array of info stored related to the solver

    ETD Parameters (see ETDAS class in etd module)
    ______________
    modecutoff : float
    contour_points :  int
    contour_radius : float 

    """
    
    def __init__(self,linop : np.ndarray,NLfunc : Callable[[np.ndarray],np.ndarray],\
            modecutoff : float = 0.01, contour_points : int = 32,contour_radius : float = 1.0):
        """
        INPUTS
        ______

        linop : np.array
            Linear operator (L) in the system dtU = LU + NL(U). Can be either a 2D numpy array (matrix)
            or a 1D array (diagonal system). L can be either real-valued or complex-valued.

        NLfunc : function 
            Nonlinear function (NL(U)) in the system dtU = LU + NL(U). Can be a complex or real-valued function.

        diagonalize : bool, optional
            Diagonalize the linear operator (matrix) and solve the diagonalized system 

        ETDCS variables: modecutoff, contour_points, contour_radius (see ETDAS documentation from etd module)

        """
        super().__init__(linop,NLfunc,modecutoff=modecutoff,\
                contour_points=contour_points,contour_radius=contour_radius)
        self._method = None
        if self._diag:
            self._method = _ETD4_Diagonal(linop,NLfunc,self.contour_points,\
                    self.contour_radius,self.modecutoff)
        else:
            self._method = _ETD4_NonDiagonal(linop,NLfunc,self.contour_points,\
                    self.contour_radius)
        self.__N1_init = False

    def _reset(self):
        """ Resets solver to its initial state """
        self.__N1_init = False
        self._h_coeff = None

    def _updateCoeffs(self,h):
        """ Update coefficients if step size h changed """
        if h == self._h_coeff:
            return
        self._h_coeff = h
        self._method._updateCoeffs(h)
        self.logs.append("ETD4 coefficients updated")
        
    def _updateStages(self,u,h):
        """ Computes u_{n+1} from u_{n} in one RK passthrough """
        self._updateCoeffs(h)
        if not self.__N1_init:
            self._method._N1_init(u)
            self.__N1_init = True
        return self._method._updateStages(u)





from rkstiff.grids import construct_x_kx_rfft, construct_x_Dx_cheb
from rkstiff.etd34 import ETD34
from rkstiff.etd35 import ETD35
from rkstiff.if34 import IF34
from rkstiff.if45dp import IF45DP
from rkstiff.etd import ETDAS
from rkstiff.etd4 import ETD4
from rkstiff.etd5 import ETD5
from rkstiff.if4 import IF4
from rkstiff.solver import StiffSolverAS
import rkstiff.models as models
import numpy as np
import pytest

# Check that soliton 'shape' is preserved through propagation
def NLstub(u):
    return 

def Lstub(N):
    return np.ones(N)

def test_solver_input():
    # Test that invalid linear operator raises ValueError
    with pytest.raises(ValueError):
        L = np.zeros(shape=(4,2))
        solver = ETD34(linop=L,NLfunc=NLstub)
    # Test that invalid linear increment factor raises ValueError
    with pytest.raises(ValueError):
        solver = ETD34(linop = Lstub(10),NLfunc=NLstub,incrF = 0.8)
    # Test that invalid linear decrement factor raises ValueError
    with pytest.raises(ValueError):
        solver = ETD34(linop = Lstub(10),NLfunc=NLstub,decrF = 1.1)
    # Test that negative epsilon raises ValueError
    with pytest.raises(ValueError):
        solver = ETD34(linop = Lstub(10),NLfunc=NLstub,epsilon = -1e-3)
    # Test that invalid safetyFactor raises ValueError 
    with pytest.raises(ValueError):
        solver = ETD34(linop = Lstub(10),NLfunc=NLstub,safetyF = 1.2)
    # Test that invalid adapt_cutoff raises ValueError
    with pytest.raises(ValueError):
        solver = ETD34(linop = Lstub(10),NLfunc=NLstub,adapt_cutoff = 1)
    with pytest.raises(ValueError):
        solver = ETD34(linop = Lstub(10),NLfunc=NLstub,minh = 0)

def test_etdsolver_input():
    # Test that modecutoff is valid
    with pytest.raises(ValueError):
        solver = ETD35(linop = Lstub(10),NLfunc=NLstub,modecutoff=1.2)
    with pytest.raises(ValueError):
        solver = ETD35(linop = Lstub(10),NLfunc=NLstub,modecutoff=0)
    # Test that contour_points is an integer
    with pytest.raises(TypeError):
        solver = ETD35(linop = Lstub(10),NLfunc=NLstub,contour_points = 12.2)
    # Test that contour_points is greater than 1
    with pytest.raises(ValueError):
        solver = ETD35(linop = Lstub(10),NLfunc=NLstub,contour_points = 1)
    # Test that contour_radius is positive
    with pytest.raises(ValueError):
        solver = ETD35(linop = Lstub(10),NLfunc=NLstub,contour_radius = 0)

def test_abc_error():
    # Test that abstract base classes cannot be instantiated
    with pytest.raises(TypeError):
        solver = ETDAS(linop = Lstub(10),NLfunc=NLstub)
    with pytest.raises(TypeError):
        solver = StiffSolverAS(linop = Lstub(10),NLfunc=NLstub)


def allen_cahn_setup():
    N = 20 
    a = -1; b = 1
    x,Dx = construct_x_Dx_cheb(N,a,b)
    epsilon = 0.01
    L,NL = models.allenCahnOps(x,Dx,epsilon)
    u0 = 0.53*x + 0.47*np.sin(-1.5*np.pi*x)
    w0 = u0 - x
    w0int = w0[1:-1]; u0int = u0[1:-1]; xint = x[1:-1]
    return xint,u0int,w0int,L,NL

def burgers_setup():
    N = 1024
    a,b = -np.pi, np.pi
    x,kx = construct_x_kx_rfft(N,a,b)

    mu = 0.0005
    L,NL = models.burgersOps(kx,mu)

    u0 = np.exp(-10*np.sin(x/2)**2)
    u0FFT = np.fft.rfft(u0)
    return u0FFT,L,NL

def kdv_soliton_setup():
    N = 256
    a,b = -30,30
    x,kx = construct_x_kx_rfft(N,a,b)
    A, x0, t0 = 1., -5., 0

    h = 0.025
    steps = 200
    tf = h*steps

    u0 = models.kdvSoliton(x,A=A,x0=x0,t=t0)
    u0FFT = np.fft.rfft(u0)
    uexactFFT = np.fft.rfft(models.kdvSoliton(x,A=A,x0=x0,t=tf))
    L,NL = models.kdvOps(kx)

    return u0FFT,L,NL,uexactFFT,h,steps

def kdv_CS_step_eval(solver,u0FFT,uexactFFT,h,steps,tol):
    for _ in range(steps):
        u0FFT = solver.step(u0FFT,h)
    rel_err = np.linalg.norm(u0FFT-uexactFFT)/np.linalg.norm(uexactFFT)
    assert rel_err < tol 

def kdv_AS_step_eval(solver,u0FFT,uexactFFT,h,steps,tol):
    for _ in range(steps):
        u0FFT,h_actual,h_suggest = solver.step(u0FFT,h)
        assert (h_actual - h) < 1e-10
    rel_err = np.linalg.norm(u0FFT-uexactFFT)/np.linalg.norm(uexactFFT)
    assert rel_err < tol 

def kdv_evolve_eval(solver,u0FFT,uexactFFT,h,tf,tol):
    uFFT = solver.evolve(u0FFT,0.0,tf,h,store_data=False)
    rel_err = np.linalg.norm(uFFT-uexactFFT)/np.linalg.norm(uexactFFT)
    assert rel_err < tol 

def test_etd34_nondiag():
    xint,u0int,w0int,L,NL = allen_cahn_setup()
    solver = ETD34(linop=L,NLfunc=NL,epsilon=1e-3,contour_points=64,contour_radius=20)
    wfint = solver.evolve(w0int,t0=0,tf=60,store_data=False)
    ufint = wfint.real + xint
    assert np.abs(u0int[0]-ufint[0]) < 0.01
    assert np.abs(u0int[7]-ufint[7]) > 1

def test_etd34():
    u0FFT,L,NL,uexactFFT,h,steps = kdv_soliton_setup()
    solver = ETD34(linop=L,NLfunc=NL,epsilon=1e-1) # small epsilon -> step size isn't auto-reduced
    kdv_AS_step_eval(solver,u0FFT,uexactFFT,h,steps,tol=1e-4)
    solver.reset()
    solver.epsilon = 1e-4
    kdv_evolve_eval(solver,u0FFT,uexactFFT,h,tf=h*steps,tol=1e-4)

def test_etd35():
    u0FFT,L,NL,uexactFFT,h,steps = kdv_soliton_setup()
    solver = ETD35(linop=L,NLfunc=NL,epsilon=1e-1) # small epsilon -> step size isn't auto-reduced
    kdv_AS_step_eval(solver,u0FFT,uexactFFT,h,steps,tol=1e-4)
    solver.reset()
    solver.epsilon = 1e-4
    kdv_evolve_eval(solver,u0FFT,uexactFFT,h,tf=h*steps,tol=1e-5)

def test_etd35_nondiag():
    xint,u0int,w0int,L,NL = allen_cahn_setup()
    solver = ETD35(linop=L,NLfunc=NL,epsilon=1e-4,contour_points=32,contour_radius=10)
    wfint = solver.evolve(w0int,t0=0,tf=60,store_data=False)
    ufint = wfint.real + xint
    assert np.abs(u0int[0]-ufint[0]) < 0.01
    assert np.abs(u0int[7]-ufint[7]) > 1

def test_if34():
    u0FFT,L,NL = burgers_setup()
    solver = IF34(linop=L,NLfunc=NL,epsilon=1e-4)
    uFFT = solver.evolve(u0FFT,t0=0,tf=0.85,store_data=False)
    rel_err = np.abs(np.linalg.norm(uFFT)-np.linalg.norm(u0FFT))/np.linalg.norm(u0FFT) 
    assert rel_err < 1e-2

def test_if34_nondiag():
    xint,u0int,w0int,L,NL = allen_cahn_setup()
    solver = IF34(linop=L,NLfunc=NL,epsilon=1e-3)
    wfint = solver.evolve(w0int,t0=0,tf=60,store_data=False)
    ufint = wfint.real + xint
    assert np.abs(u0int[0]-ufint[0]) < 0.01
    assert np.abs(u0int[7]-ufint[7]) > 1

def test_if45dp():
    u0FFT,L,NL = burgers_setup()
    solver = IF45DP(linop=L,NLfunc=NL,epsilon=1e-3)
    uFFT = solver.evolve(u0FFT,t0=0,tf=0.85,store_data=False)
    rel_err = np.abs(np.linalg.norm(uFFT)-np.linalg.norm(u0FFT))/np.linalg.norm(u0FFT) 
    assert rel_err < 1e-2


def test_etd4_step():
    u0FFT,L,NL,uexactFFT,h,steps = kdv_soliton_setup()
    solver = ETD4(linop=L,NLfunc=NL)
    kdv_CS_step_eval(solver,u0FFT,uexactFFT,h,steps,1e-6)

def test_etd4_evolve():
    u0FFT,L,NL,uexactFFT,h,steps = kdv_soliton_setup()
    solver = ETD4(linop=L,NLfunc=NL)
    kdv_evolve_eval(solver,u0FFT,uexactFFT,h,tf=h*steps,tol=1e-6)

def test_etd5_step():
    u0FFT,L,NL,uexactFFT,h,steps = kdv_soliton_setup()
    solver = ETD5(linop=L,NLfunc=NL)
    kdv_CS_step_eval(solver,u0FFT,uexactFFT,h,steps,tol=1e-6)

def test_etd5_evolve():
    u0FFT,L,NL,uexactFFT,h,steps = kdv_soliton_setup()
    solver = ETD5(linop=L,NLfunc=NL)
    kdv_evolve_eval(solver,u0FFT,uexactFFT,h=h,tf=h*steps,tol=1e-6)

def test_if4_step():
    u0FFT,L,NL,uexactFFT,h,steps = kdv_soliton_setup()
    solver = IF4(linop=L,NLfunc=NL)
    kdv_CS_step_eval(solver,u0FFT,uexactFFT,h,steps,tol=1e-5)

def test_if4_evolve():
    u0FFT,L,NL,uexactFFT,h,steps = kdv_soliton_setup()
    solver = IF4(linop=L,NLfunc=NL)
    kdv_evolve_eval(solver,u0FFT,uexactFFT,h=h,tf=h*steps,tol=1e-5)






from rkstiff.solver import StiffSolverCS
import numpy as np
from typing import Callable
from scipy.linalg import expm

class _IF4_Diagonal:
    """ IF4 diagonal system strategy for IF4 solver """
    def __init__(self,linop,NLfunc):
        self.linop = linop
        self.NLfunc = NLfunc
        
        N = linop.shape[0]
        self._EL, self._EL2 = [np.zeros(N,dtype=np.complex128) for _ in range(2)]
        self._NL1, self._NL2, self._NL3,self._NL4 = [np.zeros(N,dtype=np.complex128) for _ in range(4)]
        self._k = np.zeros(N,dtype=np.complex128) 
        
    def _updateCoeffs(self,h):
        z = h*self.linop
        self._EL = np.exp(z)
        self._EL2 = np.exp(z/2)
    
    def _N1_init(self,u):
        """ Need to initialize N1 before first updateStage call """
        self._NL1 = self.NLfunc(u)
        
    def _updateStages(self,u,h):        
        """ One RK step """
        self._k = self._EL2*u + h*self._EL2*self._NL1/2.0
        self._NL2 = self.NLfunc(self._k)
        self._k = self._EL2*u + h*self._NL2/2.
        self._NL3 = self.NLfunc(self._k)
        self._k = self._EL*u + h*self._EL2*self._NL3
        self._NL4 = self.NLfunc(self._k)
        self._k = self._EL*u + h*(self._EL*self._NL1/6.0 + self._EL2*self._NL2/3.0 \
                + self._EL2*self._NL3/3.0 + self._NL4/6.0)
        self._NL1 = self.NLfunc(self._k) # FSAL principle
        return self._k

class _IF4_NonDiagonal:
    """ IF4 non-diagonal system strategy for IF4 solver """
    def __init__(self,linop,NLfunc):
        self.linop = linop
        self.NLfunc = NLfunc
        
        N = linop.shape[0]
        self._EL, self._EL2 = [np.zeros(shape=self.linop.shape,dtype=np.complex128) for _ in range(2)]
        self._NL1, self._NL2, self._NL3,self._NL4 = [np.zeros(N,dtype=np.complex128) for _ in range(4)]
        self._k = np.zeros(N,dtype=np.complex128) 
        
    def _updateCoeffs(self,h):   
        z = h*self.linop
        self._EL = expm(z)
        self._EL2 = expm(z/2)
        
    def _N1_init(self,u):
        self._NL1 = self.NLfunc(u)
        
    def _updateStages(self,u,h):        
        self._k = self._EL2.dot(u) + h*self._EL2.dot(self._NL1/2.0)
        self._NL2 = self.NLfunc(self._k)
        self._k = self._EL2.dot(u) + h*self._NL2/2.
        self._NL3 = self.NLfunc(self._k)
        self._k = self._EL.dot(u) + h*self._EL2.dot(self._NL3)
        self._NL4 = self.NLfunc(self._k)
        self._k = self._EL.dot(u) + h*(self._EL.dot(self._NL1/6.0) + self._EL2.dot(self._NL2/3.0) \
                + self._EL2.dot(self._NL3/3.0) + self._NL4/6.0)
        self._NL1 = self.NLfunc(self._k) # FSAL principle
        return self._k

class IF4(StiffSolverCS):
    """
    Integrating factor constant step solver of 4th order 

    ATTRIBUTES
    __________
    linop : np.array
    NLfunc : function
    t : time-array stored with evolve function call 
    u : output-array stored with evolve function call 
    logs : array of info stored related to the solver

    StiffSolverAS Parameters (see StiffSolverAS class in solver module)
    ________________________
    epsilon : float
    incrF : float
    decrF : float 
    safetyF : float
    adapt_cutoff : float
    minh : float 
    """

    def __init__(self,linop : np.ndarray,NLfunc : Callable[[np.ndarray],np.ndarray]):
        """
        INPUTS
        ______

        linop : np.array
            Linear operator (L) in the system dtU = LU + NL(U). Can be either a 2D numpy array (matrix)
            or a 1D array (diagonal system). L can be either real-valued or complex-valued.

        NLfunc : function 
            Nonlinear function (NL(U)) in the system dtU = LU + NL(U). Can be a complex or real-valued function.

        """

        super().__init__(linop,NLfunc)
        self._method = None
        if self._diag:
            self._method = _IF4_Diagonal(linop,NLfunc)
        else:
            self._method = _IF4_NonDiagonal(linop,NLfunc)
        self._reset()

    def _reset(self):
        """ Resets solver to its initial state  """
        self.__N1_init = False
        self._h_coeff = None
        
    def _updateCoeffs(self,h):
        """ Update coefficients if step size h changed """
        if h == self._h_coeff:
            return
        self._h_coeff = h
        self._method._updateCoeffs(h)
        self.logs.append("IF4 coefficients updated")
        
    def _updateStages(self,u,h):
        """ Computes u_{n+1} from u_{n} through one RK passthrough """
        self._updateCoeffs(h)
        if not self.__N1_init:
            self._method._N1_init(u)
            self.__N1_init = True
        return self._method._updateStages(u,h)
    
